import os
import pytest

from cybele.aes.cipher import passphrase_to_key
from cybele.api.core import Database

HERE = os.path.dirname(__file__)


@pytest.fixture
def db():
    passphrase = "passphrase"
    key = passphrase_to_key(passphrase)
    path_to_db = os.path.join(HERE, "..", "data", "encrypted_fake_entries.json")
    db = Database(path_to_db, key)
    return db



import os
import tempfile

import pytest

from cybele.aes import passphrase_to_key
from cybele.aes.exception import WrongPassphraseError
from cybele.api.core import Database
from cybele.api import exceptions

from .fixtures import db

HERE = os.path.dirname(__file__)


def test_raise_no_such_db():
    with pytest.raises(exceptions.DBNotFoundError):
        _ = Database("fake/path/to/db", key=None)


def test_raise_when_wrong_passphrase():
    passphrase = "wrong_passphrase"
    key = passphrase_to_key(passphrase)
    path_to_db = os.path.join(HERE, "..", "data", "encrypted_fake_entries.json")
    with pytest.raises(WrongPassphraseError):
        db = Database(path_to_db, key)


def test_load_entries(db):
    expected_entries = {
        'entry1': {'username': 'User1', 'passphrase': 'passphrase1'},
        'entry2': {'username': 'User2', 'passphrase': 'passphrase2'}
    }
    assert db.entries == expected_entries


def test_dump_entries(db):
    with tempfile.NamedTemporaryFile() as tmp:
        db.dump(tmp.name)
        assert db.load(tmp.name) == db.entries



import json
import os

import pyperclip

from ..aes import AESCipher
from ..aes.exception import WrongPassphraseError
from .exceptions import DBNotFoundError, EntryNameNotAvailableError, EntryFormatError, EntryNotFoundError
from .utils import flush_clipboard


class Database:
    def __init__(self, path_to_db, key):
        self._check_db_exists(path_to_db)
        self.cipher = AESCipher(key)
        self.path_to_db = path_to_db

        self.entries = self.load(path_to_db)

    def add(self, entry_name, entry):
        self._check_entryname_available(entry_name)
        self._check_entry(entry)
        self.entries[entry_name] = entry

    def edit(self, entry_name, updates):
        self._check_entry_exists(entry_name)
        self.entries[entry_name].update(updates)
    
    def delete(self, entry_name):
        self._check_entry_exists(entry_name)
        del self.entries[entry_name]

    def load(self, path):
        with open(path, "r") as f:
            encrypted_data = f.read()
        try:
            entries_dumps = self.cipher.decrypt(encrypted_data)
            entries = json.loads(entries_dumps)
            return entries
        except (json.JSONDecodeError, UnicodeDecodeError) as exc:
            raise WrongPassphraseError("Wrong passphrase")

    def dump(self, path):
        entries_dumps = json.dumps(self.entries)
        encrypted_data = self.cipher.encrypt(entries_dumps)
        with open(path, "w") as f:
            f.write(encrypted_data)

    @flush_clipboard(after=10)
    def clip_passphrase(self, entry_name):
        self._check_entry_exists(entry_name)
        passphrase = self.entries[entry_name].get("passphrase")
        pyperclip.copy(passphrase)

    @flush_clipboard(after=10)
    def clip_username(self, entry_name):
        self._check_entry_exists(entry_name)
        username = self.entries[entry_name].get("username")
        pyperclip.copy(username)

    def _check_db_exists(self, path_to_db):
        if not os.path.exists(path_to_db):
            raise DBNotFoundError(f"No such DB file : '{path_to_db}'.")

    def _check_entry_exists(self, entry_name):
        if entry_name not in self.entries:
            raise EntryNotFoundError(f"Entry named '{entry_name}' can't be found in DB.")

    def _check_entryname_available(self, entry_name):
        if entry_name in self.entries:
            raise EntryNameNotAvailableError(f"Entry named '{entry_name}' already exists.")

    @staticmethod
    def _check_entry(entry):
        if not all(key in entry for key in ["username", "passphrase"]):
            raise EntryFormatError("Entry must have a 'username' & 'passphrase' key.")



from functools import wraps
import time

import pyperclip
from tqdm import tqdm


def flush_clipboard(after=10):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            try:
                result = f(*args, **kwargs)
                for _ in tqdm(range(after * 10), desc="Copied to clipboard", bar_format="{l_bar}|{bar}|"):
                    time.sleep(1 / 10)
                pyperclip.copy("")
                return result
            finally:
                pyperclip.copy("")

        return wrapper

    return decorator



import base64
import hashlib
import os

from Crypto.Cipher import AES


def passphrase_to_key(passphrase):
    return hashlib.sha256(passphrase.encode("utf8")).digest()


class AESCipher(object):
    def __init__(self, key):
        self.iv_size = 16
        self.key = key

    def encrypt(self, plaintext):
        plaintext = self._pad(plaintext)
        iv = os.urandom(self.iv_size)
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        encrypted = cipher.encrypt(plaintext)
        return base64.b64encode(iv + encrypted).decode('utf8')

    def decrypt(self, encrypted):
        encrypted = base64.b64decode(encrypted.encode('utf8'))
        iv, encrypted = encrypted[:self.iv_size], encrypted[self.iv_size:]
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return self._unpad(cipher.decrypt(encrypted)).decode('utf8')

    def _pad(self, s):
        s = s.encode('utf8')
        return s + (self.iv_size - len(s) % self.iv_size) * chr(self.iv_size - len(s) % self.iv_size).encode('utf8')

    @staticmethod
    def _unpad(s):
        return s[:-ord(s[len(s) - 1:])]



from cybele.aes import AESCipher, passphrase_to_key


def test_encrypt_decrypt():
    plaintext = "plaintext"
    passphrase = "passphrase"
    key = passphrase_to_key(passphrase)
    cipher1 = AESCipher(key)
    cipher2 = AESCipher(key)
    encrypted = cipher1.encrypt(plaintext)
    decrypted = cipher2.decrypt(encrypted)
    assert decrypted == plaintext


def test_different_encryption_for_same_plaintext():
    plaintext = "plaintext"
    passphrase = "passphrase"
    key = passphrase_to_key(passphrase)
    cipher1 = AESCipher(key)
    cipher2 = AESCipher(key)
    encrypted1 = cipher1.encrypt(plaintext)
    encrypted2 = cipher2.encrypt(plaintext)
    assert encrypted1 != encrypted2



import json
import os
import sys

import click
from terminaltables import SingleTable

from ..aes import AESCipher, passphrase_to_key
from ..aes.exception import WrongPassphraseError
from ..api import Database


class Table(SingleTable):
    def __init__(self, data, **kwargs):
        SingleTable.__init__(self, data)
        for attr, value in kwargs.items():
            if not hasattr(self, attr):
                raise AttributeError(f"Table has no '{attr}' attribute")
            setattr(self, attr, value)

    def __str__(self):
        return str(self.table)


def load_db(db_path, msg="Enter passphrase"):
    try:
        key = passphrase_to_key(ask_passphrase(msg=msg))
        db = Database(db_path, key)
        return db
    except WrongPassphraseError as exc:
        echo_exit("Wrong passphrase !", 1)


def confirm_yes_no(msg):
    choice = click.prompt(f"{msg} (yes/no)")
    if choice not in ["y", "n", "yes", "no"]:
        choice = confirm_yes_no(msg)
    if choice in ["n", "no"]:
        return False
    if choice in ["y", "yes"]:
        return True


def echo_exit(msg, exit_code=1):
    click.echo(msg)
    sys.exit(exit_code)


def ask_passphrase(msg="Enter passphrase"):
    return click.prompt(msg, hide_input=True)


def ask_and_confirm_passphrase(msg="Enter passphrase"):
    passphrase = click.prompt(msg, hide_input=True)
    confirmation = click.prompt("Enter same passphrase again", hide_input=True)
    if passphrase != confirmation:
        echo_exit("Passphrases do not match. Try again")
    return passphrase


def ask_full_infos():
    username = ask_username()
    passphrase = ask_and_confirm_passphrase()
    return {"username": username, "passphrase": passphrase}


def ask_username():
    return click.prompt("Username")


def create_db(db_path):
    db_dir, db_name = os.path.split(db_path)
    if not os.path.exists(db_dir):
        os.makedirs(db_dir)
    if os.path.exists(db_path):
        echo_exit(f"DB {db_path} already exists.")

    passphrase = ask_and_confirm_passphrase(msg="Enter main passphrase")
    key = passphrase_to_key(passphrase)
    cipher = AESCipher(key)
    with open(db_path, "w") as f:
        dumps = json.dumps({})
        f.write(cipher.encrypt(dumps))



import os

import pytest
import pyperclip

from cybele.api import exceptions

from .fixtures import db

HERE = os.path.dirname(__file__)


def test_add_entry(db):
    db.add("new_entry", {"username": "new_username", "passphrase": "new_passphrase"})
    assert db.entries.get("new_entry") == {"username": "new_username", "passphrase": "new_passphrase"}


def test_edit_entry(db):
    db.edit("entry1", {"passphrase": "edited_passphrase"})
    assert db.entries.get("entry1") == {"username": "User1", "passphrase": "edited_passphrase"}


def test_edit_delete(db):
    db.delete("entry1")
    assert "entry1" not in db.entries


def test_raise_when_unavailable_on_add(db):
    with pytest.raises(exceptions.EntryNameNotAvailableError):
        db.add("entry1", {"username": "new_username", "passphrase": "new_passphrase"})


def test_raise_when_format_ko_on_add(db):
    with pytest.raises(exceptions.EntryFormatError):
        db.add("new_entry", {"not_good_key": "new_username", "passphrase": "new_passphrase"})


def test_raise_when_entry_not_found_on_edit(db):
    with pytest.raises(exceptions.EntryNotFoundError):
        db.edit("wrong_entry_name", {})


def test_raise_when_entry_not_found_on_delete(db):
    with pytest.raises(exceptions.EntryNotFoundError):
        db.delete("wrong_entry_name")



import click

from ..aes import AESCipher
from ..constants import DEFAULT_DB_PATH, USER_HOME
from . import utils


@click.group()
@click.option("--db", default=DEFAULT_DB_PATH, type=click.Path(),
              help=f"path to db (default: {DEFAULT_DB_PATH})".replace(USER_HOME, "~"))
@click.pass_context
def cybele(ctx, db):
    """
    command line password manager
    """
    ctx.obj = {
        "db_path": db,
    }


@cybele.command()
@click.pass_context
def chgpass(ctx):
    """change db password"""
    db_path = ctx.obj["db_path"]
    db = utils.load_db(db_path, msg="Enter old passphrase")
    new_passphrase = utils.ask_and_confirm_passphrase(msg="Enter new passphrase")
    key = utils.passphrase_to_key(new_passphrase)
    db.cipher = AESCipher(key)
    db.dump(db_path)


@cybele.command()
@click.option("--db", default=DEFAULT_DB_PATH, type=click.Path(),
              help=f"path to db (default: {DEFAULT_DB_PATH})".replace(USER_HOME, "~"))
def init(db):
    """initialize a new db"""
    utils.create_db(db)


@cybele.command()
@click.option("-g", "--grep", default="", help="filter entries matching string")
@click.pass_context
def ls(ctx, grep):
    """list all entries"""
    db_path = ctx.obj["db_path"]
    db = utils.load_db(db_path)
    if not db.entries:
        utils.echo_exit("No entries yet. See 'cybele add --help'", 0)

    headers = [["ENTRY", "USERNAME", "PASSPHRASE"]]
    data = headers + [
        [entry_name, entry["username"], "*" * 10]
        for entry_name, entry in db.entries.items()
        if grep in entry_name or grep in entry["username"]
    ]
    table = utils.Table(data, inner_heading_row_border=True)
    click.echo(f"{table}")


@cybele.command()
@click.argument("entry_name")
@click.pass_context
def add(ctx, entry_name):
    """add new entry"""
    db_path = ctx.obj["db_path"]
    db = utils.load_db(db_path)
    entry = utils.ask_full_infos()
    db.add(entry_name=entry_name, entry=entry)
    db.dump(db_path)


@cybele.command()
@click.argument("entry_name")
@click.pass_context
def edit(ctx, entry_name):
    """edit an entry"""
    db_path = ctx.obj["db_path"]
    db = utils.load_db(db_path)
    entry = utils.ask_full_infos()
    db.edit(entry_name=entry_name, updates=entry)
    db.dump(db_path)


@cybele.command()
@click.argument("entry_name")
@click.pass_context
def rm(ctx, entry_name):
    """removes an entry"""
    db_path = ctx.obj["db_path"]
    db = utils.load_db(db_path)
    confirmation = utils.confirm_yes_no(f"Are you sure you want to delete '{entry_name}' ?")
    if confirmation:
        db.delete(entry_name)
    db.dump(db_path)


@cybele.command()
@click.argument("entry_name")
@click.option('-u', "--username", is_flag=True, help="copy entry username instead of passphrase")
@click.pass_context
def cp(ctx, entry_name, username):
    """copy passphrase (or username) to clipboard"""
    db_path = ctx.obj["db_path"]
    db = utils.load_db(db_path)
    if username:
        db.clip_username(entry_name)
    else:
        db.clip_passphrase(entry_name)



class DBNotFoundError(FileNotFoundError):
    """Raise when DB is not reachable"""


class EntryNotFoundError(KeyError):
    """Raise when trying to set entry_name and its not free"""


class EntryNameNotAvailableError(KeyError):
    """Raise when trying to set entry_name and its not free"""


class EntryFormatError(ValueError):
    """Raise when core keys are missing in a record"""



from unittest.mock import patch

import pyperclip
import pytest

from .fixtures import db


def pyperclip_available():
    try:
        pyperclip.paste()
        return False
    except pyperclip.PyperclipException as exc:
        return True


@pytest.mark.skipif(pyperclip_available(), reason="pyperclip not available on system")
def test_pyperclip_dependency():
    pyperclip.copy("to be copied")
    assert pyperclip.paste() == "to be copied"


@pytest.mark.skipif(pyperclip_available(), reason="pyperclip not available on system")
@patch("cybele.api.utils.tqdm", return_value=[])
def test_flush_on_clip_username(tqdm_mock, db):
    pyperclip.copy("supposed to be erased")
    db.clip_username("entry1")
    assert pyperclip.paste() == ""


@pytest.mark.skipif(pyperclip_available(), reason="pyperclip not available on system")
@patch("cybele.api.utils.tqdm", return_value=[])
def test_flush_on_clip_passphrase(tqdm_mock, db):
    pyperclip.copy("supposed to be erased")
    db.clip_passphrase("entry1")
    assert pyperclip.paste() == ""


@pytest.mark.skipif(pyperclip_available(), reason="pyperclip not available on system")
@patch("cybele.api.utils.tqdm", return_value=range(10))
@patch("cybele.api.utils.time.sleep", side_effect=Exception())
def test_flush_even_if_exception(sleep_mock, tqdm_mock, db):
    pyperclip.copy("supposed to be erased")
    with pytest.raises(Exception):
        db.clip_username("entry1")
    assert pyperclip.paste() == ""



from setuptools import setup, find_packages

with open("README.md") as f:
    readme = f.read()

setup(
    name="cybele",
    version="0.3.1",
    packages=find_packages(),
    include_package_data=True,
    description="cli password manager",
    long_description=readme,
    long_description_content_type="text/markdown",
    install_requires=[
        "Click==7.0",
        "pycrypto==2.6.1",
        "pyperclip==1.7.0",
        "terminaltables==3.1.0",
        "tqdm==4.31.1"
    ],
    url="https://gitlab.com/dithyrambe/cybele",
    classifiers=[
        "Programming Language :: Python :: 3.6",
    ],
    license="MIT",
    entry_points={
        'console_scripts': [
            "cybele = cybele.cli:cybele"
        ]
    }
)
